# BSD LICENSE
#
# Copyright (c) 2014, Ixia All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#   Version 6.20
#   
#   File: ixFileUtils.tcl
#    Copyright   IXIA
#    All Rights Reserved.
#
#    Revision Log:
#    09-05-2002    MG    Genesis
#
# Description: Contains miscellaneous file commands.
#
##################################################################################


namespace eval ixFileUtils {
    variable filesToClose ""
}


########################################################################
# Procedure:   ixFileUtils::closeOpenFiles
#
# Description: Either close all open file handles or close all file handles opened by Ixia routines
#              The advanced test parameter closeAllFilesInCleanUp controls which one is done
#
# Arguments:   None
#
# Returns:     Nothing
########################################################################
proc ixFileUtils::closeAll {} \
{
    variable filesToClose

    if {[info tclversion] >= 8.3} {
        set closeAllFiles [advancedTestParameter cget -closeAllFilesInCleanUp]
        set listOfFiles ""
        if {$closeAllFiles} {
            set listOfFiles $filesToClose
        } else {
            foreach fileId [file channels file*] {
                set index [lsearch -exact $filesToClose $fileId]
                if {$index >= 0} {
                    lappend listOfFiles $fileId
                }
            }
        }
        foreach fileId $listOfFiles {
            close $fileId
        }        
    }
}


########################################################################
# Procedure:   ixFileUtils::addFileToList
#
# Description: Add a file handle to the list of handles to be closed during the cleanUp procedure
#
# Arguments:   fileId - the file handle to add
#
# Returns:     Nothing
########################################################################
proc ixFileUtils::addFileToList { fileId } \
{
    variable filesToClose

    lappend filesToClose $fileId
}


########################################################################
# Procedure:   ixFileUtils::removeFileFromList
#
# Description: Remove a file handle from the list of handles to be closed during the cleanUp procedure
#
# Arguments:   fileId - the file handle to remove
#
# Returns:     Nothing
########################################################################
proc ixFileUtils::removeFileFromList { fileId } \
{
    variable filesToClose

    set index [lsearch -exact $filesToClose $fileId]
    if {$index >= 0} {
        set filesToClose [lreplace $filesToClose $index $index]
    }
}


########################################################################
# Procedure:   ixFileUtils::open
#
# Description: This command opens a file
#
# Arguments:   filename - full path to filename
#              access - file access parameter to use
#              permission - file permission parameter to use
#
# Returns:     fileID if successful, or "" if unsuccessful
########################################################################
proc ixFileUtils::open {filename {access ""} {permission ""}} \
{
    if {[catch {eval ::open [list $filename] $access $permission} fileID]} {
        set fileID ""
    } else {
        ixFileUtils::addFileToList $fileID
    }
    return $fileID
}


########################################################################
# Procedure:   ixFileUtils::close
#
# Description: This command closes a file
#
# Arguments:   fileId - channel id of the file to be closed
#
# Returns:     Nothing
########################################################################
proc ixFileUtils::close {fileId} \
{
    catch {::close $fileId}
    ixFileUtils::removeFileFromList $fileId
}


