# BSD LICENSE
#
# Copyright (c) 2014, Ixia All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#   Version 6.20
#   
#   File: vlanUtils.tcl
#
#   Copyright   IXIA
#   All Rights Reserved.
#
#	Revision Log:
#	11-02-2001	DS
#
# Description: This file contains common vlan utilities
#
########################################################################

proc vlanUtilsSetDefault {} {
    set ::vlanUtils::untaggedPortList ""
}


namespace eval vlanUtils {} {
    vlanUtilsSetDefault
}


########################################################################
# Procedure: vlanUtils::setPortTagged
#
# Description: Sets a boolean indicated whether this port is a vlan-tagged port
#
# Arguments:
#   chassis - The number of the chassis
#   card - The number of the card
#   port - The number of the port
#
# Returns: TCL_OK on completion
########################################################################
proc vlanUtils::setPortTagged {chassis card port} \
{
    variable untaggedPortList

    set retCode $::TCL_OK

    if {[catch {lsearch $untaggedPortList [list $chassis $card $port]} found]} {
        set found -1
    }

    if {$found >= 0} {
        set untaggedPortList [lreplace $untaggedPortList $found $found]
    }

    return $retCode
}


########################################################################
# Procedure: vlanUtils::setTagged
#
# Description: Sets a boolean indicated whether the ports in this list are vlan-tagged ports
#
# Arguments: portList - A list of ports to tag.  A port identifier is of the list form: chassis card port.
#
# Returns: TCL_OK on success and TCL_ERROR on failure
########################################################################
proc vlanUtils::setTagged {portList} \
{
    set retCode $::TCL_OK

    foreach taggedPort $portList {
        scan $taggedPort "%d %d %d" c l p

        if {[setPortTagged $c $l $p]} {
            set retCode $::TCL_ERROR
            break
        }
    }

    return $retCode
}


########################################################################
# Procedure: vlanUtils::setPortUntagged
#
# Description: Sets a boolean indicated whether this port is a vlan-tagged port
#
# Arguments:
#   chassis - The number of the chassis for the port
#   card - The number of the card for the port
#   port - The number of the port
#
# Returns: TCL_OK on success
########################################################################
proc vlanUtils::setPortUntagged {chassis card port} \
{
    variable untaggedPortList

    set retCode $::TCL_OK

    if {[catch {lsearch $untaggedPortList [list $chassis $card $port]} found]} {
        set found -1
    }

    if {$found < 0} {
        lappend untaggedPortList [list $chassis $card $port]
    }

    return $retCode
}


########################################################################
# Procedure: vlanUtils::setUntagged
#
# Description: Sets a boolean indicated whether the ports in this list are NOT vlan-tagged ports
#
# Arguments: portList - A list of ports to tag.  A port identifier is of the list form: chassis card port.
#
# Returns: TCL_OK on success and TCL_ERROR on failure
########################################################################
proc vlanUtils::setUntagged {portList} \
{
    set retCode $::TCL_OK

    foreach taggedPort $portList {
        scan $taggedPort "%d %d %d" c l p

        if {[setPortUntagged $c $l $p]} {
            set retCode $::TCL_ERROR
            break
        }
    }

    return $retCode
}


########################################################################
# Procedure: vlanUtils::isPortTagged
#
# Description: Returns a boolean indicated whether this port is a vlan-tagged port
#
# Arguments:
#   chassis - The number of the chassis for the port
#   card - The number of the card for the port
#   port - The number of the port
#
# Returns: true if the port is tagged and false if it is not.
########################################################################
proc vlanUtils::isPortTagged {chassis card port} \
{
    variable untaggedPortList

    if {[catch {lsearch $untaggedPortList [list $chassis $card $port]} found]} {
        set tagged $::true
    } else {

        if {$found < 0} {
            set tagged $::true
        } else {
            set tagged $::false
        }
    }

    return $tagged
}


########################################################################
# Procedure: emptyUntaggedPortList
#
# Description: Reset the untagged port list to a empty list.
#
# Arguments:
#
# Returns:
########################################################################
proc vlanUtils::emptyUntaggedPortList {} \
{
    variable untaggedPortList

    set untaggedPortList ""
}


########################################################################
# Procedure: setPortTagged
#
# Description: Sets a boolean indicated whether this port is a vlan-tagged port
#
# Arguments:
#   chassis - The number of the chassis
#   card - The number of the card
#   port - The number of the port
#
# Returns: TCL_OK on completion
########################################################################
proc setPortTagged {chassis card port} \
{
    return [vlanUtils::setPortTagged $chassis $card $port]
}


########################################################################
# Procedure: setTagged
#
# Description: Sets a boolean indicated whether the ports in this list are vlan-tagged ports
#
# Arguments: portList - A list of ports to tag.  A port identifier is of the list form: chassis card port.
#
# Returns: TCL_OK on success and TCL_ERROR on failure
########################################################################
proc setTagged {portList} \
{
    return [vlanUtils::setTagged $portList]
}


########################################################################
# Procedure: setPortUntagged
#
# Description: Sets a boolean indicated whether this port is a vlan-tagged port
#
# Arguments:
#   chassis - The number of the chassis for the port
#   card - The number of the card for the port
#   port - The number of the port
#
# Returns: TCL_OK on success
########################################################################
proc setPortUntagged {chassis card port} \
{
    return [vlanUtils::setPortUntagged $chassis $card $port]
}


########################################################################
# Procedure: setUntagged
#
# Description: Sets a boolean indicated whether the ports in this list are NOT vlan-tagged ports
#
# Arguments: portList - A list of ports to tag.  A port identifier is of the list form: chassis card port.
#
# Returns: TCL_OK on success and TCL_ERROR on failure
########################################################################
proc setUntagged {portList} \
{
    return [vlanUtils::setUntagged $portList]
}


########################################################################
# Procedure: isPortTagged
#
# Description: Returns a boolean indicated whether this port is a vlan-tagged port
#
# Arguments:
#   chassis - The number of the chassis for the port
#   card - The number of the card for the port
#   port - The number of the port
#
# Returns: true if the port is tagged and false if it is not.
########################################################################
proc isPortTagged {chassis card port} \
{
    return [vlanUtils::isPortTagged $chassis $card $port]
}


########################################################################
# Procedure: getUntaggedPortList
#
# Description: Returns the list of untagged ports
#
# Arguments:
#
# Returns: The list of untagged ports.
########################################################################
proc getUntaggedPortList {} \
{
    return $::vlanUtils::untaggedPortList
}


########################################################################
# Procedure: emptyUntaggedPortList
#
# Description: Reset the untagged port list to a empty list.
#
# Arguments:
#
# Returns:
########################################################################
proc emptyUntaggedPortList {} \
{
    vlanUtils::emptyUntaggedPortList
}

